"""Collection of helper functions for OA package

@author: Pieter Eendebak <pieter.eendebak@gmail.com>
"""

# %% Load packages

import fileinput
import functools
import inspect
import logging
import operator
import os
import re
import subprocess
import sys
import tempfile
import time
import traceback
import warnings
import webbrowser
from time import gmtime, strftime
from typing import Any, Literal

import dateutil.parser
import numpy as np

try:
    import matplotlib
    import matplotlib.pyplot as plt
except BaseException:
    warnings.warn("oahelper: matplotlib cannot be found, not all functionality is available")

import oapackage

try:
    import oalib
except ImportError:
    warnings.warn("oahelper: oalib cannot be found, not all functionality is available")
from oapackage import markup


def deprecated(func):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used."""

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        try:
            filename = inspect.getfile(func)
        except BaseException:
            filename = "?"
        try:
            lineno = inspect.getlineno(func)
        except BaseException:
            lineno = -1
        warnings.warn_explicit(
            f"Call to deprecated function {func.__name__}.",
            category=UserWarning,
            filename=filename,
            lineno=lineno,
        )
        return func(*args, **kwargs)

    return new_func


# %%


try:
    try:
        import qtpy.QtGui
        import qtpy.QtWidgets
    except BaseException:
        # no Qt support
        pass

    _applocalqt = None

    def monitorSizes(verbose=0):
        """Return monitor sizes"""
        if sys.platform == "win32":
            import ctypes

            user32 = ctypes.windll.user32
            wa = [[0, 0, user32.GetSystemMetrics(0), user32.GetSystemMetrics(1)]]
        else:
            _applocalqt = qtpy.QtWidgets.QApplication.instance()
            if _applocalqt is None:
                _applocalqt = qtpy.QtWidgets.QApplication([])
                _qd = qtpy.QtWidgets.QDesktopWidget()
            else:
                _qd = qtpy.QtWidgets.QDesktopWidget()

            nmon = _qd.screenCount()
            wa = [_qd.screenGeometry(ii) for ii in range(nmon)]
            wa = [[w.x(), w.y(), w.width(), w.height()] for w in wa]

            if verbose:
                for ii, w in enumerate(wa):
                    print("monitor %d: %s" % (ii, str(w)))
        return wa

except BaseException:

    def monitorSizes(verbose=0):
        """Dummy implementation"""
        return [[0, 0, 1280, 720]]


def guitest_monitorSizes():
    sizes = monitorSizes()
    assert isinstance(sizes, list)


def tilefigs(lst, geometry, ww=None, raisewindows=False, tofront=False, verbose=0):
    """Tile figure windows on a specified area"""
    mngr = plt.get_current_fig_manager()
    be = matplotlib.get_backend()
    if ww is None:
        ww = monitorSizes()[-1]

    w = ww[2] / geometry[0]
    h = ww[3] / geometry[1]

    if verbose:
        print(f"tilefigs: ww {str(ww)}, w {w} h {h}")
    for ii, f in enumerate(lst):
        if not plt.fignum_exists(f):
            continue
        fig = plt.figure(f)
        iim = ii % np.prod(geometry)
        ix = iim % geometry[0]
        iy = np.floor(float(iim) / geometry[0])
        x = ww[0] + ix * w
        y = ww[1] + iy * h
        if verbose:
            print(f"ii {ii}: {ix} {iy}: f {f}: {x} {y} {w} {h}")
            if verbose >= 2:
                print(f"  window {mngr.get_window_title()}")
        if be == "WXAgg":
            fig.canvas.manager.window.SetPosition((x, y))
            fig.canvas.manager.window.SetSize((w, h))
        if be == "WX":
            fig.canvas.manager.window.SetPosition((x, y))
            fig.canvas.manager.window.SetSize((w, h))
        if be == "agg":
            fig.canvas.manager.window.SetPosition((x, y))
            fig.canvas.manager.window.resize(w, h)
        if be == "Qt4Agg" or be == "QT4" or be == "QT5Agg":
            # assume Qt canvas
            try:
                fig.canvas.manager.window.move(x, y)
                fig.canvas.manager.window.resize(w, h)
                fig.canvas.manager.window.setGeometry(x, y, w, h)
                # mngr.window.setGeometry(x,y,w,h)
            except Exception as ex:
                print(
                    "problem with window manager: ",
                )
                print(f"backend {be}")
                logging.exception(ex)
        if raisewindows:
            mngr.window.raise_()
        if tofront:
            plt.figure(f)


def plot2Dline(line, *args, **kwargs):
    """Plot a 2D line in a matplotlib figure"""
    if np.abs(line[1]) > 0.001:
        xx = plt.xlim()
        xx = np.array(xx)
        yy = (-line[2] - line[0] * xx) / line[1]
        plt.plot(xx, yy, *args, **kwargs)
    else:
        yy = np.array(plt.ylim())
        xx = (-line[2] - line[1] * yy) / line[0]
        plt.plot(xx, yy, *args, **kwargs)


# %% Make nice plots


def niceplot(
    ax, fig=None, despine=True, verbose=0, figurebg=True, tightlayout=True, legend=None, almost_black="#222222"
):
    """Create a good looking plot

    The code:
        - removes spines
        - makes legend and spines lighter
        - makes legend lighter

    """

    if verbose:
        print("niceplot: remove spines")

    spines_to_keep = ["bottom", "left"]
    if despine:
        spines_to_remove = ["top", "right"]
        for spine in spines_to_remove:
            ax.spines[spine].set_visible(False)
    else:
        spines_to_keep += ["top", "right"]
    ax.get_xaxis().tick_bottom()
    ax.get_yaxis().tick_left()

    if verbose:
        print("niceplot: reduce spine intensity")

    for spine in spines_to_keep:
        ax.spines[spine].set_linewidth(0.5)
        ax.spines[spine].set_color(almost_black)

    ax.tick_params(axis="both", direction="out")

    if legend is not None:
        if verbose:
            print("niceplot: adjust legend")

        # Remove the line around the legend box, and instead fill it with a light grey
        # Also only use one point for the scatterplot legend because the user will
        # get the idea after just one, they don't need three.
        light_grey = np.array([float(241) / float(255)] * 3)
        rect = legend.get_frame()
        rect.set_facecolor(light_grey)
        middle_grey = np.array([float(151) / float(255)] * 3)
        rect.set_edgecolor(middle_grey)
        # rect.set_linewidth(0.0)

        # Change the legend label colors to almost black, too
        texts = legend.texts
        for t in texts:
            t.set_color(almost_black)

    if (fig is not None) and tightlayout:
        fig.tight_layout(pad=1.0)

    if figurebg and fig is not None:
        fig.set_facecolor("w")

    plt.draw()
    plt.show()


def enlargelims(factor: float = 1.05):
    """Enlarge the limits of a plot

    Args:
        factor (float): factor by which to make the plot margins wider
    """
    xl = plt.xlim()
    d = (factor - 1) * (xl[1] - xl[0]) / 2
    xl = (xl[0] - d, xl[1] + d)
    plt.xlim(xl)
    yl = plt.ylim()
    d = (factor - 1) * (yl[1] - yl[0]) / 2
    yl = (yl[0] - d, yl[1] + d)
    plt.ylim(yl)


# %%


def helmert_contrasts(number_of_levels, verbose=0):
    """Calculate Helmert contrasts for a given number of levels

    The Helmert contrasts are orthogonal and normalize such that the square equals to number of levels.
    Args:
        number_of_levels: number of levels in the number_of_levels
    Returns:
        array: array with calculated Helmert contrasts
    """
    N = number_of_levels
    meoffset = 0
    md = number_of_levels - 1

    main_effects = np.zeros((number_of_levels, md))
    Z = np.zeros((number_of_levels, md + 1))

    for value in range(number_of_levels):
        for ii in range(0, md + 1):
            Z[value, ii] = value > ii - 1

        # make Helmert contrasts (these are automatically orthogonal)
        Z[value, 0] = 1
        if value > 0:
            Z[value, value] = value

        for q in range(1, value):
            Z[value, q] = 0
        for q in range(value + 1, md + 1):
            Z[value, q] = -1

    if verbose:
        print(f"helmert_contrasts: {number_of_levels}\n")
        print("Z (initial creation)")
        print(Z)

    # normalize the contrasts
    for ii in range(0, md):
        normalization = Z[:, (ii + 1)].T.dot(Z[:, ii + 1])
        if verbose:
            print(f"helmert_contrasts: normalize number_of_levels tmp: {normalization} ")
        main_effects[:, meoffset + ii : (meoffset + ii + 1)] = (
            np.sqrt(N) * Z[:, (ii + 1) : (ii + 2)] / np.sqrt(normalization)
        )

    return main_effects


def selectParetoArrays(array_list, pareto_object):
    """Select arrays using a Pareto object

    Args:
        array_list (list): list of arrays
        pareto (object): oapackage Pareto object
    Returns:
        list: list with all Pareto optimal designs
    """
    paretoarrays = oalib.arraylist_t()
    paretoidx = np.array(pareto_object.allindices())
    ww = oalib.longVector(tuple(paretoidx.tolist()))
    oalib.selectArrays(array_list, ww, paretoarrays)
    return paretoarrays


def joinArrayLists(ww):
    """Concetenate a list of array lists into a single list"""
    ll = oalib.arraylist_t()
    for w in ww:
        for al in w:
            ll.push_back(al)
    return ll


def createPareto(dds, verbose=1):
    """Create Pareto object from dataset"""
    pp = oapackage.ParetoDoubleLong()

    for ii in range(dds.shape[0]):
        v = (dds[ii, :]).astype(np.float64)
        pp.addvalue(v, ii)

    if verbose:
        pp.show(verbose)
    return pp


# %% Utils


def static_var(varname: str, value: Any):
    """Helper function to create a static variable"""

    def decorate(func):
        setattr(func, varname, value)
        return func

    return decorate


@static_var("time", 0)
def tprint(string, dt=1, output=False):
    """Print progress of a loop every dt seconds"""
    if (time.time() - tprint.time) > dt:
        print(string)
        tprint.time = time.time()
        if output:
            return True
        else:
            return
    else:
        if output:
            return False
        else:
            return


def timeString(tt=None) -> str:
    """Return a string with the current time or specified time

    Args:
        tt (struct_time or None): time to convert
    Returns:
        Formatted time
    """
    if tt is None:
        tt = gmtime()
    ts = strftime("%Y-%m-%d %H:%M:%S", tt)
    return ts


def findfilesR(directory: str, pattern) -> list[str]:
    """Get a list of files (recursive)

    Args:
        directory (str): directory to search
        patt (str): pattern
    Returns:
        list: list of matched files
    """
    lst = []
    for root, _, files in os.walk(directory, topdown=False):
        lst += [os.path.join(root, f) for f in files]
    rr = re.compile(pattern)
    lst = [filename for filename in lst if re.match(rr, filename)]
    return lst


def findfiles(directory: str, pattern: str | None = None) -> list[str]:
    """Get a list of files
    Args:
        directory: directory to search
        patt : pattern
    Returns:
        List of matched files
    """
    lst = os.listdir(directory)
    if pattern is not None:
        rr = re.compile(pattern)
        lst = [filename for filename in lst if re.match(rr, filename)]
    return lst


def finddirectories(directory: str, pattern: str | None = None) -> list[str]:
    """Get a list of directories
    Args:
        directory: directory to search
        patt: pattern
    Returns:
        List of matched directories
    """
    lst = os.listdir(directory)
    if pattern is not None:
        rr = re.compile(pattern)
        lst = [filename for filename in lst if re.match(rr, filename)]
    lst = [filename for filename in lst if os.path.isdir(os.path.join(directory, filename))]
    return lst


def oainfo(filename: str, verbose: int = 1):
    """Print information about a file containing arrays"""
    af = oapackage.arrayfile_t(filename, verbose)
    print(af.showstr())
    af.closefile()


def oaIsBinary(filename: str) -> bool:
    """Return true if array file is in binary format"""
    af = oapackage.arrayfile_t(filename)
    ret = af.isbinary()
    af.closefile()
    return ret


def fac(n: int) -> int:
    """Return n! (factorial)

    Args:
        n (int): number to calculate factorial
    Returns
        integer: factorial of number
    """
    if n == 1 or n == 0:
        return 1
    else:
        return n * fac(n - 1)


def choose(n: int, k: int) -> int:
    """Return n choose k"""
    ntok = 1
    for t in range(min(k, n - k)):
        ntok = ntok * (n - t) // (t + 1)
    return ntok


def array2latex(
    X,
    header: bool = True,
    hlines=(),
    floatfmt: str = "%g",
    comment: str | None = None,
    hlinespace: None | float = None,
    mode: Literal["tabular", "psmallmatrix", "pmatrix"] = "tabular",
    tabchar: str = "c",
) -> str:
    """Convert numpy array to Latex tabular or matrix"""
    ss = ""
    if comment is not None:
        if isinstance(comment, list):
            for line in comment:
                ss += f"% {str(line)}\n"
        else:
            ss += f"% {str(comment)}\n"
    if header:
        match mode:
            case "tabular":
                if len(tabchar) == 1:
                    cc = tabchar * X.shape[1]
                else:
                    cc = tabchar + tabchar[-1] * (X.shape[1] - len(tabchar))
                ss += f"\\begin{{tabular}}{{{cc}}}" + chr(10)
            case "psmallmatrix":
                ss += "\\begin{psmallmatrix}" + chr(10)
            case "pmatrix":
                ss += "\\begin{pmatrix}" + chr(10)
            case _:
                raise ValueError(f"mode {mode} is invalid")
    for ii in range(X.shape[0]):
        r = X[ii, :]
        if isinstance(r[0], str):
            ss += " & ".join([f"{x}" for x in r])
        else:
            ss += " & ".join([floatfmt % x for x in r])
        if ii < (X.shape[0]) - 1 or not header:
            ss += "  \\\\" + chr(10)
        else:
            ss += "  " + chr(10)
        if ii in hlines:
            ss += r"\hline" + chr(10)
            if hlinespace is not None:
                ss += f"\\rule[+{hlinespace:.2f}ex]{{0pt}}{{0pt}}"
    if header:
        match mode:
            case "tabular":
                ss += "\\end{tabular}"
            case "psmallmatrix":
                ss += "\\end{psmallmatrix}" + chr(10)
            case "pmatrix":
                ss += "\\end{pmatrix}" + chr(10)
            case _:
                raise ValueError(f"mode {mode} is invalid")
    return ss


def array2html(X, header=1, tablestyle="border-collapse: collapse;", trclass="", tdstyle="", trstyle="", thstyle=""):
    """Convert Numpy array to HTML table

    Arguments
    ---------
        X : numpy array
            array to be converted
        header : integer
            use header or not
    Returns
    -------
        page : markup html object
            generated table in HTML

    """
    page = markup.page()
    page.add("<!-- Created by array2html -->\n")
    page.table(style=tablestyle)
    offset = 0
    nc = X.shape[1]
    nr = X.shape[0]

    if isinstance(trstyle, str):
        trstyle = [trstyle] * nr
    if isinstance(trclass, str):
        trclass = [trclass] * nr

    ri = 0
    if header:
        page.tr(style="font-weight: bold; border-bottom: solid 1px black;" + trstyle[ri], class_=trclass[ri])
        ri = ri + 1
        for ii in range(nc):
            if isinstance(X[offset, ii], tuple):
                print("array2html: tuple instance")
                page.th(X[offset, ii][0], style=thstyle + X[offset, ii][1])
            else:
                page.th(X[offset, ii], style=thstyle)
        page.tr.close()
        offset = offset + 1

    nr = X.shape[0] - offset
    for _ in range(nr):
        page.tr(style=trstyle[ri], _class=trclass[ri])
        for ii in range(nc):
            if isinstance(X[offset, ii], tuple):
                page.td(X[offset, ii][0], style=tdstyle + X[offset, ii][1])
            else:
                page.td(X[offset, ii], style=tdstyle)

        page.tr.close()
        offset = offset + 1
        ri = ri + 1
    page.table.close()
    return page


def write_text_arrayfile(filename: str, designs: list[Any], comment: str | None = None):
    """Write designs to disk in text format"""
    nrows = designs[0].n_rows
    ncols = designs[0].n_columns
    afile = oapackage.oalib.arrayfile_t(filename, nrows, ncols, len(designs), oapackage.oalib.ATEXT, 8)
    if comment is not None:
        for c in comment.split("\n"):
            afile.add_comment(c)
    afile.append_arrays(designs)
    afile.closefile()


def arrayfile_generator(afile: str):
    """Return generator to read all files in the array file"""
    af = oapackage.arrayfile_t(afile)

    na = af.narrays
    for ii in range(na):
        yield af.readnext()
    af.closefile()


def runcommand(
    cmd: str,
    dryrun=0,
    idstr: str | None = None,
    verbose: int = 1,
    logfile: str | None = None,
    shell: bool = True,
):
    """Run specified command in external environment

    Args:
        cmd: Command to execute

    Returns:
        r (int): return value of the shell command
    """
    if idstr is not None:
        cmd = f'echo "idstr: {idstr}";\n' + cmd
    if verbose >= 2:
        print(f"cmd: {cmd}")
    r: int | None = 0
    if not dryrun:
        process = subprocess.Popen(cmd, bufsize=1, stdout=subprocess.PIPE, shell=shell)
        for jj in range(10000000):
            r = process.poll()
            assert process.stdout is not None
            line_bytes = process.stdout.readline()
            line = line_bytes.decode(encoding="UTF-8")
            if verbose >= 2:
                print(f"runcommand: jj {jj}")
            if verbose >= 3:
                print(f'runcommand: jj {jj}: "{line}"')
            if len(line) == 0:
                if verbose >= 2:
                    print(f"runcommand: len(line) {len(line)}")

                break

            if r is not None:
                break
            if jj > 20000000:
                print("error: early abort of runcommand")
                break
            sys.stdout.write(str(line))
            if jj % 2 == 0:
                sys.stdout.flush()
            if verbose >= 2:
                print("end of loop...")
        r = process.poll()
        if not r == 0:
            print(f"runcommand: cmd returned error! r={str(r)}")
            print(cmd)
            return r
    else:
        if verbose >= 2 or (verbose and logfile is None):
            print(f"### dryrun of {idstr}\n{cmd}\n###\n")
    if logfile is not None:
        fid = open(logfile, "a")
        fid.write("\n###\n")
        fid.write(cmd)
        fid.close()
    else:
        pass

    return r


def getArrayFile(afile: str) -> str:
    """Return pointer to array file
    Automatically check for compressed array files
    """
    afilegz = afile = afile + ".gz"
    if not os.path.exists(afile) and os.path.exists(afilegz):
        afile = afilegz
    else:
        afile = afile
    return afile


def checkOAfile(afile, verbose=0):
    """Return pointer to array file

    Automatically check for compressed array files

    Args:
        afile (str): location of array file
    Returns:
        arrayfile_t or None
    """
    af = oapackage.arrayfile_t(afile, verbose)
    nm = af.filename
    if af.isopen():
        return nm
    else:
        return None


def checkArrayFile(afile, cache=1):
    """Check whether a file or list of files exists
    If the file does not exist, check whether a compressed file does exist
    cache: 0 (no), 1 (check), -1 (always)
    """
    if cache == -1:
        return True
    if cache == 0:
        return False

    if os.path.exists(afile):
        return True
    else:
        if afile.endswith(".oa"):
            if os.path.exists(afile + ".gz"):
                return True
            else:
                return False
        else:
            return False


def isstr(s: Any) -> bool:
    return isinstance(s, str)


def checkFiles(lst, cache=1, verbose=0):
    """Check whether a file or list of files exists
    cache: 0 (no), 1 (check), -1 (always)

    Returns False if one or more of the files do not exist
    Returns True if all files exist
    """
    if cache == -1:
        return True
    if cache == 0:
        return False

    if isstr(lst):
        lst = [lst]
    for f in lst:
        if not os.path.exists(f):
            if verbose:
                print(f"checkFiles: file {f} does not exist")
            return False
    return True


def test_checkFiles():
    def touch(fname):
        if os.path.exists(fname):
            os.utime(fname, None)
        else:
            open(fname, "a").close()

    lst = [tempfile.mktemp()]  # ty: ignore
    r = checkFiles(lst, cache=1, verbose=1)
    assert r is False
    touch(lst[0])
    r = checkFiles(lst, cache=1, verbose=1)
    assert r is True


def checkFilesOA(lst, cache=1, verbose=0):
    """Check whether a file or list of files exists

    Args:
        lst (list): list of files
        cache (int): 0 (always return False), 1 (check), -1 (always return True)

    For array files also the .gz extension is checked
    Returns False if one or more of the files do not exist
    Returns True if all files exist
    """
    if verbose >= 2:
        print(f"checkFilesOA: cache {cache}")
    if cache == -1:
        return True
    if cache == 0:
        return False

    if isstr(lst):
        lst = [lst]
    c = True
    for f in lst:
        r = checkArrayFile(f, cache=True)
        if not r:
            if verbose:
                print(f"checkFiles: file {f} does not exist")
            c = False
            break
    return c


def randomizearrayfile(input_filename, output_filename, verbose=1):
    """Randomize a file with orthogonal arrays

    Each array is transformed with a random transformation

    Args:
        afile (str): input file
        afileout (str): output file
    """
    lst = oapackage.readarrayfile(input_filename)
    rlst = oapackage.oalib.arraylist_t()
    for ii, al in enumerate(lst):
        adata = oalib.arraylink2arraydata(al)
        trans = oalib.array_transformation_t(adata)
        if verbose >= 2:
            print("randomizing array %d\n" % ii)
        trans.randomize()
        alr = trans.apply(al)
        rlst.push_back(alr)
    oapackage.writearrayfile(output_filename, rlst)


def nArrayFile(filename, verbose=1):
    """Return number of arrays in file

    Args:
        afile (str): name of array file
    """
    af = oalib.arrayfile_t(filename, 0)
    n = af.narrays
    af.closefile()
    return n


def selectArraysInFile(infile, outfile, idx, afmode=oalib.ATEXT, verbose=1, cache=1):
    """Select arrays in a file by indices

    Args:
        infile (str): file with designs
        outfile (str): output  file with designs
        inx (list of indices)
    Returns:
        None

    """
    if not checkFiles(outfile, cache=cache):
        gidxint = oalib.intVector([int(x) for x in idx])
        sols = oalib.arraylist_t()
        oalib.selectArrays(infile, gidxint, sols, 0)
        af = oalib.arrayfile_t(infile, 1)
        _ = oalib.writearrayfile(outfile, sols, afmode, af.nrows, af.ncols)
        if verbose >= 2:
            print("selectArrays: write array file %s in mode %d" % (outfile, afmode))
    else:
        if verbose >= 2:
            print(f"output file {outfile} already exists")


selectArrays = deprecated(selectArraysInFile)


def floatformat(number, mind=2, maxd=4):
    """Format a floating point number into a string"""
    rounded = int(number * (10**mind)) / float(10**mind)
    if number == rounded:
        return ("%%.0%df" % mind) % number
    else:
        return ("%%.0%dg" % maxd) % number


def safemin(data, default=0):
    """Return mininum of array with default value for empty array

    Args:
        data (array or list): data to return the maximum
        default (obj): default value
    Returns:
        object: minimum value in the array or the default value

    """
    if isinstance(data, list):
        if len(data) == 0:
            minimum_value = default
        else:
            minimum_value = min(data)
        return minimum_value
    if data.size == 0:
        minimum_value = default
    else:
        minimum_value = data.min()
    return minimum_value


def safemax(data, default=0):
    """Return maximum of array with default value for empty array

    Args:
        data (array or list ): data to return the maximum
        default (obj): default value
    Returns:
        object: maximum value in the array or the default value
    """

    if isinstance(data, list):
        if len(data) == 0:
            maximum_value = default
        else:
            maximum_value = max(data)
        return maximum_value
    if data.size == 0:
        maximum_value = default
    else:
        maximum_value = data.max()
    return maximum_value


def mkdirc(directory_name: str) -> str:
    """Create directory"""
    if not os.path.exists(directory_name):
        os.mkdir(directory_name)
    return directory_name


def parseProcessingTime(logfile, verbose=0):
    """Parse a log file to calculate the processing time"""

    fileinput.close()
    tstart = None
    tend = None
    dtr = None
    try:
        for line in fileinput.input([logfile]):
            if line.startswith("#time"):
                if verbose >= 1:
                    print(f"parseProcessingTime: line: {line}", end="")
            if line.startswith("#time start:"):
                tstart = dateutil.parser.parse(line[13:])
                if verbose >= 2:
                    print(f"parseProcessingTime: tstart: {tstart}")
            elif line.startswith("#time end:"):
                tend = dateutil.parser.parse(line[10:])
                if verbose >= 2:
                    print(f"parseProcessingTime: tend: {tend}")
            elif line.startswith("#time total:"):
                dtr = float(line[13:][:-5])
                if verbose >= 2:
                    print(f"parseProcessingTime: total: {dtr}")
            else:
                pass
        if tstart is not None and tend is not None:
            dt = tend - tstart
            dtt = dt.total_seconds()
        else:
            dtt = -1
    except BaseException:
        if verbose:
            print(f"error processing log {logfile}")
            traceback.print_exc(file=sys.stdout)
        dtt = -1
    if dtr is not None:
        if abs(dtr - dtt) > 10:
            print(f"parseProcessingTime: warning difference in reported times {dtt:.1f} dtr {dtr:.1f} [s]")
    return dtt


def series2htmlstr(ad, html=1, case=0):
    """Convert arraydata_t to html formatted string"""
    s = list(ad.factor_levels())
    p = -1
    n = 0
    levels = []
    bb = []
    while n < len(s):
        if s[n] != p:
            p = s[n]
            levels.append(p)
            bb.append(1)
        else:
            bb[-1] += 1
        n = n + 1
    if case == 0:
        if bb[-1] > 1:
            bb[-1] = "a"
    hstr = f"OA({ad.N}; {ad.strength}; "
    for ii, _ in enumerate(levels):
        if html:
            hstr += f"{levels[ii]}<sup>{bb[ii]}</sup>"
        else:
            hstr += f"{levels[ii]}^{bb[ii]}"
    hstr += ")"
    return hstr


def gwlp2str(gmadata, t=None, sformat=None, jstr=","):
    """Convert GWLP value to string format"""
    if gmadata is None:
        return "-"
    if isinstance(gmadata, tuple):
        pass
    else:
        if isinstance(gmadata, list):
            pass
        else:
            gmadata[gmadata < 0] = 0
        if not (np.abs(gmadata[0] - 1) < 1e-12 and np.abs(gmadata[1]) < 1e-12):
            warnings.warn("data does not represent GWPL data", UserWarning)
            return ""
    bgma = np.around(gmadata, decimals=12)
    if t is not None:
        bgma = bgma[(t + 1) :]
    if sformat is None:
        gstr = jstr.join([floatformat(v, mind=2, maxd=4) for v in bgma])
    else:
        gstr = jstr.join([sformat % v for v in bgma])
    return gstr


def selectJ(sols0, jj=5, jresults=None, verbose=1):
    """Select only arrays with J-characteristics non-zero

    We asssume the designs are in even-odd ordering (i.e. only check the J value of first columns)
    """
    if jresults is None:
        jresults = oalib.analyseArrays(sols0, verbose, jj)

    solseo = oalib.arraylist_t()
    v = []
    for ii, js in enumerate(jresults):
        v.append(js.values[0])

    si = [i for (i, j) in sorted(enumerate(v), key=operator.itemgetter(1))]
    for jj in si:
        if v[jj] > 0:
            solseo.append(sols0[jj])
    if verbose:
        print(f"selectJ: kept {solseo.size()}/{len(sols0)} solutions")
    return solseo


def test_selectJ():
    al = oapackage.exampleArray(12, 1)
    sols0 = [al]
    r = selectJ(sols0)
    assert len(r) == 0


def extendSingleArray(A, adata, t=3, verbose=1):
    """Extend a single orthogonal array"""
    oaoptions = oalib.OAextend()
    oaoptions.setAlgorithmAuto(adata)
    sols0 = oalib.arraylist_t()
    solsx = oalib.arraylist_t()
    sols0.push_back(A)
    k = A.n_columns
    n = oalib.extend_arraylist(sols0, adata, oaoptions, k, solsx)
    assert n >= len(solsx)
    sys.stdout.flush()
    return solsx


def runExtend(N: int, k: int, t: int = 3, l: int = 2, verbose: int = 1, initsols=None, nums=[], algorithm=None):  # noqa
    """Run extension algorithm and return arrays

    Args:
      N (int): number of rows
      k (int): number of columns to extend to
      t (int): strength of the arrays
      l (int): factors of the designs
      initsols (None or list): list of arrays to extend, None to start with root

    Returns:
        list: list of generated designs

    Example:
       >>> import oapackage
       >>> designs = oapackage.oahelper.runExtend(16, 5, 3, verbose=0)
    """
    if verbose:
        print(f"runExtend: N={N}, k={k}, t={t}")
    if isinstance(l, list):
        ll = l
    else:
        ll = [l]
    ll = ll + [ll[-1]] * (k - len(ll))
    s = oalib.intVector(ll)
    adata = oalib.arraydata_t(s, N, t, k)
    al = oalib.array_link(adata.N, adata.strength, 1)
    al.create_root(adata)
    if initsols is None:
        sols0 = oalib.arraylist_t()
        sols0.append(al)
        tstart = t
    else:
        sols0 = initsols
        tstart = sols0[0].n_columns

    oaoptions = oalib.OAextend()
    if algorithm is None:
        oaoptions.setAlgorithmAuto(adata)
    else:
        oaoptions.setAlgorithm(algorithm, adata)
    solsx = sols0
    for ii in range(tstart, k):
        solsx = oalib.arraylist_t()
        oalib.extend_arraylist(sols0, adata, oaoptions, ii, solsx)
        if verbose >= 2:
            print(f" ii {ii}: {solsx.size()}")
        sols0 = solsx
        nums.append(solsx.size())
        sys.stdout.flush()
    return solsx


def compressOAfile(afile, decompress=False, verbose=1):
    """Compress an OA array file

    Args:
        afile (str): array to compress
        decompress (bool): If True then decompress
        verbose (int): verbosity level
    """
    af = oalib.arrayfile_t(afile, 0)
    if decompress:
        raise NotImplementedError("decompressing file not implemted")

    if verbose >= 2:
        print(f"file {afile}: binary {af.isbinary()}")
    if not (sys.platform == "linux2" or sys.platform == "linux"):
        if verbose:
            print(f"compressOAfile: not compressing file {afile} (platform not supported)")
        return False
    if af.isbinary() and not af.iscompressed and sys.platform == "linux2":
        if verbose:
            print(f"compressOAfile: compressing file {afile}")
        cmd = f"gzip -f {afile}"
        r = os.system(cmd)
        if not r == 0:
            print(f"compressOAfile: error compressing file {afile}")
        return True
    else:
        if not af.isbinary():
            if verbose:
                print(f"compressOAfile: not compressing file {afile} (file is in text mode)")
        else:
            if verbose:
                print(f"compressOAfile: not compressing file {afile} (file {af.filename} is compressed)")
        return False


def argsort(seq):
    """Stable version of argsort"""
    # http://stackoverflow.com/questions/3382352/equivalent-of-numpy-argsort-in-basic-python/3382369#3382369
    return sorted(range(len(seq)), key=seq.__getitem__)


@deprecated
def jseq(xx, comb):
    pp = functools.reduce(lambda y, i: xx[:, i] + y, comb, 0)
    jseq = 1 - 2 * np.mod(pp, 2)
    return jseq


def sortrows(x):
    """Sort rows of an array, return indices

    The array is sorted low to high with the first column as the major column.

    Args:
        array (numpy array)
    Returns:
        list: indices of rows of sorted array
    """
    if len(x.shape) == 1:
        nn = 1
        sind = np.argsort(x)
        return sind
    else:
        nn = x.shape[1]
    dt = [("x%d" % xx, "f8") for xx in range(0, nn)]
    if x.size == 0:
        sind = np.array([])
        return sind
    v = [tuple(x[kk, :]) for kk in range(0, x.shape[0])]
    w = np.array(v, dtype=dt)
    sind = np.argsort(w)
    return sind


def sortcols(array):
    """Sort columns of an array, return indices

    The array is sorted low to high with the row column as the major row.

    Args:
        array (numpy array)
    Returns:
        list: indices of columns of sorted array
    """
    sind = sortrows(array.transpose())
    return sind


# %%


def testHtml(html_code=None):
    """Test a short snippet of HTML"""
    if html_code is None:
        return
    page = markup.page()
    page.init()
    page.body()
    page.add(html_code)
    page.body.close()
    _, f = tempfile.mkstemp(".html")
    with open(f, "w") as fname:
        fname.write(str(page))
        fname.close()
        webbrowser.open(fname.name)


def designStandardError(al) -> tuple[float, float, float]:
    """Return standard errors for a design

    Args:
      al (array): design

    Returns:
        array: array with standard errors

    """

    X = np.array(al.getModelMatrix(2))
    k = al.n_columns

    scalefac = 1
    M = np.linalg.inv(X.transpose().dot(X) / scalefac)

    mm = np.array(M.diagonal()).flatten()

    m1 = mm[1 : (1 + k)].flatten()
    m1 = m1[np.argsort(m1)]
    m2 = mm[(1 + k) :]
    m2 = m2[np.argsort(m2)]
    m0 = mm[0]
    return np.sqrt(m0), np.sqrt(m1), np.sqrt(m2)  # ty: ignore


# %%
def DefficiencyBound(D: float, k: int, k2: int) -> float:
    """Calculate the D-efficiency bound of an array extension

    Args:
        D: D-efficiency of the design
        k: numbers of columns
        k2: numbers of columns in extension

    Returns:
        Bound on the D-efficiency of extensions of a design with k columns to k2 columns

    """
    m = 1.0 + k + k * (k - 1) / 2
    m2 = 1.0 + k2 + k2 * (k2 - 1) / 2
    Dbound = D ** (m / m2)
    return Dbound


# %% Misc


def setWindowRectangle(x, y=None, w=None, h=None, mngr=None, be=None):
    """Position the current Matplotlib figure at the specified position"""
    if y is None:
        y = x[1]
        w = x[2]
        h = x[3]
        x = x[0]
    if mngr is None:
        mngr = plt.get_current_fig_manager()
    be = matplotlib.get_backend()
    window = mngr.canvas.manager.window
    if be == "WXAgg":
        window.SetPosition((x, y))
        window.SetSize((w, h))
    elif be == "agg":
        window.SetPosition((x, y))
        window.resize(w, h)
    elif be == "module://IPython.kernel.zmq.pylab.backend_inline":
        pass
    else:
        # assume Qt canvas
        window.move(x, y)
        window.resize(w, h)
        window.setGeometry(x, y, w, h)


def makearraylink(array):
    """Convert array to array_link object

    Args:
        array (numpy array or array_link): array to convert
    Returns:
        array_link
    """
    if isinstance(array, np.ndarray):
        array = oapackage.create_array_link(array)
    return array


def formatC(al, wrap: bool = True) -> str:
    """Format array for inclusion in C code

    Args:
        array: Array to be converted to C code representation
        wrap: If True, then add code to create a variable
    Returns:
        Formatted string
    """
    A = np.array(al).T.flatten().tolist()
    s = ",".join(["%d" % x for x in A])
    if wrap:
        s = "\tarray_link array ( %d,%d, 0 );\n\tint array_data_tmp[] = {" % (al.n_rows, al.n_columns) + s + "};"
        s += "\tarray.setarraydata(array_data_tmp, array.n_rows * array.n_columns);\n"
    return s


def create_pareto_element(values, pareto=None):
    """Create a vector of mvalue_t elements
    Args:
        vv (list): list with tuples or arrays
    Returns:
        object: Pareto element
    """
    if pareto is None:
        pareto = oalib.ParetoDoubleLong()
    if isinstance(pareto, oalib.ParetoMultiLongLong):
        vector_pareto = oalib.mvalueVector()
        for v in values:
            if isinstance(v, (int)):
                # convert to list type
                v = [v]
            if not isinstance(v, (list, type)):
                raise Exception(
                    f"creating Pareto element for Pareto object of type {type(pareto)} "
                    f"and input of type {type(v)} not supported"
                )
            vec = oalib.mvalue_t_long(list(v))
            vector_pareto.push_back(vec)
    elif isinstance(pareto, oalib.ParetoMultiDoubleLong):
        vector_pareto = oalib.vector_mvalue_t_double()
        for v in values:
            if isinstance(v, (int, float)):
                # convert to list type
                v = [float(v)]
            if not isinstance(v, (list, tuple)):
                raise Exception(
                    f"creating Pareto element for Pareto object of type {type(pareto)} "
                    f"and input of type {type(v)} not supported"
                )

            vec = oalib.mvalue_t_double(list(v))
            vector_pareto.push_back(vec)
    elif isinstance(pareto, oalib.ParetoDoubleLong):
        if not isinstance(values, (list, tuple, np.ndarray)):
            raise Exception(f"cannot handle input of type {tuple(values)}")
        vector_pareto = values
    else:
        raise Exception(
            f"creating Pareto element for Pareto object of type {type(pareto)} "
            f"and input of type {type(values)} not supported"
        )
    return vector_pareto



"""Orthogonal Array package test functions"""

import unittest

import numpy as np

import oapackage
import oapackage.graphtools
import oapackage.scanf


class TestGraphtools(unittest.TestCase):
    """Test functionality related to orthogonal array files"""

    def test_selectIsomorphismClasses(self):
        ll = [oapackage.exampleArray(0), oapackage.exampleArray(0)]
        indices, mm = oapackage.graphtools.selectIsomorphismClasses(ll, verbose=0)
        assert indices[0] == indices[1]
        self.assertEqual(len(mm), 2)
        assert np.all(mm[0] == mm[1])


if __name__ == "__main__":
    unittest.main()



import oapackage.conference
import oapackage.Doptim
import oapackage.oahelper
from oapackage.Doptim import (  # noqa
    Doptimize,
    array2Dtable,
    calcScore,
    filterPareto,
    generateDscatter,
    optimDeffPython,
    scoreDn,
    selectDn,
)
from oapackage.oahelper import (  # noqa
    DefficiencyBound,
    array2html,
    array2latex,
    arrayfile_generator,
    checkArrayFile,
    checkFiles,
    checkFilesOA,
    checkOAfile,
    choose,
    compressOAfile,
    create_pareto_element,
    createPareto,
    deprecated,
    designStandardError,
    enlargelims,
    extendSingleArray,
    fac,
    fileinput,
    finddirectories,
    findfiles,
    findfilesR,
    floatformat,
    formatC,
    getArrayFile,
    guitest_monitorSizes,
    gwlp2str,
    helmert_contrasts,
    isstr,
    joinArrayLists,
    jseq,
    makearraylink,
    mkdirc,
    monitorSizes,
    nArrayFile,
    niceplot,
    oainfo,
    oaIsBinary,
    oalib,
    parseProcessingTime,
    plot2Dline,
    randomizearrayfile,
    runcommand,
    runExtend,
    safemax,
    safemin,
    selectArrays,
    selectArraysInFile,
    selectJ,
    selectParetoArrays,
    series2htmlstr,
    setWindowRectangle,
    sortcols,
    sortrows,
    static_var,
    strftime,
    test_selectJ,
    tilefigs,
    timeString,
    tprint,
    webbrowser,
    write_text_arrayfile,
)

""" Orthogonal Array package

The Orthogonal Array package is a package to generate and analyse orthogonal
arrays, optimal designs and conference matrices. For more information see

http://github.com/eendebakpt/oapackage

"""
from oalib import *  # noqa # legacy structure # ty: ignore
from oalib import (  # noqa
    ABINARY,
    ALATEX,
    ATEXT,
    GWLP,
    LMC_EQUAL,
    LMC_LESS,
    LMC_MORE,
    GWLPmixed,
    Jcharacteristics_conference,
    LMCcheck,
    LMCcheckOriginal,
    LMCreduction_t,
    OAextend,
    ParetoDoubleLong,
    PECsequence,
    PICsequence,
    array2eigenModelMatrixMixed,
    array2modelmatrix,
    array2modelmatrix_sizes,
    array_link,
    array_transformation_t,
    arraydata_t,
    arrayfile_t,
    arraylink2arraydata,
    arraylist_t,
    arrayrankInfo,
    compareJ54,
    compareLMC,
    conference2DSD,
    conference_design2modelmatrix,
    conference_t,
    create_array_link,
    depth_extend_t,
    distance_distribution,
    distance_distribution_mixed,
    distance_distribution_mixed_inplace,
    distance_distribution_shape,
    double_conference_foldover_permutation,
    exampleArray,
    extend_array,
    extend_conference_matrix,
    extend_conference_plain,
    extend_conference_restricted,
    generateConferenceExtensions,
    get_time_ms,
    getloglevel,
    intVector,
    is_root_form,
    isConferenceFoldover,
    jstruct_t,
    jstructconference_t,
    jvalue,
    krawtchouk,
    macwilliams_transform,
    macwilliams_transform_mixed,
    mvalue_t_double,
    mvalue_t_long,
    mycheck_handler,
    ndarray_double,
    oafilestring,
    oagraph2transformation,
    optimDeff,
    print_column,
    projDeff,
    projectionDOFvalues,
    rankStructure,
    readarrayfile,
    reduceConference,
    reduceDOPform,
    reduceGraphNauty,
    reduceLMCform,
    reduceOAnauty,
    reductionDOP,
    selectConferenceIsomorpismClasses,
    selectLMC0,
    splitDir,
    splitFile,
    splitTag,
    symmetry_group,
    throw_runtime_exception,
    transformGraphMatrix,
    update_array_link,
    writearrayfile,
)

oapackage.oalib.setloglevel(oapackage.oalib.SYSTEM)
oapackage.oalib.log_print(-oapackage.oalib.SYSTEM, "")
__version__ = oapackage.oalib.version()

__description__ = "Orthogonal Array package"
__uri__ = "http://www.pietereendebak.nl/oapackage/index.html"



#!/usr/bin/env python
"""doxy2swig.py [options] index.xml output.i

Doxygen XML to SWIG docstring converter (improved version).

Converts Doxygen generated XML files into a file containing docstrings
for use by SWIG.

index.xml is your doxygen generated XML file and output.i is where the
output will be written (the file will be clobbered).
"""
#
# The current version of this code is hosted on a github repository:
#   https://github.com/m7thon/doxy2swig
#
# This code is implemented using Mark Pilgrim's code as a guideline:
#   http://www.faqs.org/docs/diveintopython/kgp_divein.html
#
# Original Author: Prabhu Ramachandran
# Modified by:     Michael Thon (June 2015)
# License: BSD style
#
# Thanks:
#   Johan Hake:  the include_function_definition feature
#   Bill Spotz:  bug reports and testing.
#   Sebastian Henschel:   Misc. enhancements.
#
# Changes:
# June 2015 (Michael Thon):
#   - class documentation:
#     -c: add constructor call signatures and a "Constructors" section
#         collecting the respective docs (e.g. for python)
#     -a: add "Attributes" section collecting the documentation for member
#         variables (e.g. for python)
#   - overloaded functions:
#     -o: collect all documentation into one "Overloaded function" section
#   - option to include function definition / signature renamed to -f
#   - formatting:
#     + included function signatures slightly reformatted
#     + option (-t) to turn off/on type information for funciton signatures
#     + lists (incl. nested and ordered)
#     + attempt to produce docstrings that render nicely as markdown
#     + translate code, emphasis, bold, linebreak, hruler, blockquote,
#       verbatim, heading tags to markdown
#     + new text-wrapping and option -w to specify the text width
#

import optparse
import os.path
import re
import textwrap
from xml.dom import minidom


def my_open_read(source):
    if hasattr(source, "read"):
        return source
    else:
        try:
            return open(source, encoding="utf-8")
        except TypeError:
            return open(source)


def my_open_write(dest):
    if hasattr(dest, "write"):
        return dest
    else:
        try:
            return open(dest, "w", encoding="utf-8")
        except TypeError:
            return open(dest, "w")


# MARK: Text handling:


def shift(txt, indent="    ", prepend=""):
    """Return a list corresponding to the lines of text in the `txt` list
    indented by `indent`. Prepend instead the string given in `prepend` to the
    beginning of the first line. Note that if len(prepend) > len(indent), then
    `prepend` will be truncated (doing better is tricky!). This preserves a
    special '' entry at the end of `txt` (see `do_para` for the meaning).
    """
    if isinstance(indent, int):
        indent = indent * " "
    special_end = txt[-1:] == [""]
    lines = "".join(txt).splitlines(True)
    for i in range(1, len(lines)):
        if lines[i].strip() or indent.strip():
            lines[i] = indent + lines[i]
    if not lines:
        return prepend
    prepend = prepend[: len(indent)]
    indent = indent[len(prepend) :]
    lines[0] = prepend + indent + lines[0]
    ret = ["".join(lines)]
    if special_end:
        ret.append("")
    return ret


class Doxy2SWIG:
    """Converts Doxygen generated XML files into a file containing
    docstrings that can be used by SWIG-1.3.x that have support for
    feature("docstring").  Once the data is parsed it is stored in
    self.pieces.

    """

    def __init__(
        self,
        src,
        with_function_signature=False,
        with_type_info=False,
        with_constructor_list=False,
        with_attribute_list=False,
        with_overloaded_functions=False,
        textwidth=80,
        quiet=False,
    ):
        """Initialize the instance given a source object.  `src` can
        be a file or filename.  If you do not want to include function
        definitions from doxygen then set
        `include_function_definition` to `False`.  This is handy since
        this allows you to use the swig generated function definition
        using %feature("autodoc", [0,1]).

        """
        # options:
        self.with_function_signature = with_function_signature
        self.with_type_info = with_type_info
        self.with_constructor_list = with_constructor_list
        self.with_attribute_list = with_attribute_list
        self.with_overloaded_functions = with_overloaded_functions
        self.textwidth = textwidth
        self.quiet = quiet

        # state:
        self.indent = 0
        self.listitem = ""
        self.pieces = []

        f = my_open_read(src)
        self.my_dir = os.path.dirname(f.name)
        self.xmldoc = minidom.parse(f).documentElement
        f.close()

        self.pieces.append(f"\n// File: {os.path.basename(f.name)}\n")

        self.space_re = re.compile(r"\s+")
        self.lead_spc = re.compile(r'^(%feature\S+\s+\S+\s*?)"\s+(\S)')
        self.multi = 0
        self.ignores = [
            "inheritancegraph",
            "param",
            "listofallmembers",
            "innerclass",
            "name",
            "declname",
            "incdepgraph",
            "invincdepgraph",
            "programlisting",
            "type",
            "references",
            "referencedby",
            "location",
            "collaborationgraph",
            "reimplements",
            "reimplementedby",
            "derivedcompoundref",
            "basecompoundref",
            "argsstring",
            "definition",
            "exceptions",
        ]
        # self.generics = []

    def generate(self):
        """Parses the file set in the initialization.  The resulting
        data is stored in `self.pieces`.

        """
        self.parse(self.xmldoc)

    def write(self, fname):
        o = my_open_write(fname)
        o.write("".join(self.pieces))
        o.write("\n")
        o.close()

    def parse(self, node):
        """Parse a given node.  This function in turn calls the
        `parse_<nodeType>` functions which handle the respective
        nodes.

        """
        pm = getattr(self, f"parse_{node.__class__.__name__}")
        pm(node)

    def parse_Document(self, node):
        self.parse(node.documentElement)

    def parse_Text(self, node):
        txt = node.data
        if txt == " ":
            # this can happen when two tags follow in a text, e.g.,
            # " ...</emph> <formaula>$..." etc.
            # here we want to keep the space.
            self.add_text(txt)
            return
        txt = txt.replace("\\", r"\\")
        txt = txt.replace('"', r"\"")
        # ignore pure whitespace
        m = self.space_re.match(txt)
        if not (m and len(m.group()) == len(txt)):
            self.add_text(txt)

    def parse_Comment(self, node):
        """Parse a `COMMENT_NODE`.  This does nothing for now."""
        return

    def parse_Element(self, node):
        """Parse an `ELEMENT_NODE`.  This calls specific
        `do_<tagName>` handers for different elements.  If no handler
        is available the `subnode_parse` method is called.  All
        tagNames specified in `self.ignores` are simply ignored.

        """
        name = node.tagName
        ignores = self.ignores
        if name in ignores:
            return
        attr = f"do_{name}"
        if hasattr(self, attr):
            handlerMethod = getattr(self, attr)
            handlerMethod(node)
        else:
            self.subnode_parse(node)
            # if name not in self.generics: self.generics.append(name)

    # MARK: Special format parsing
    def subnode_parse(self, node, pieces=None, indent=0, ignore=[], restrict=None):
        """Parse the subnodes of a given node. Subnodes with tags in the
        `ignore` list are ignored. If pieces is given, use this as target for
        the parse results instead of self.pieces. Indent all lines by the amount
        given in `indent`. Note that the initial content in `pieces` is not
        indented. The final result is in any case added to self.pieces."""
        if pieces is not None:
            old_pieces, self.pieces = self.pieces, pieces
        else:
            old_pieces = []
        if isinstance(indent, int):
            indent = indent * " "
        if len(indent) > 0:
            pieces = "".join(self.pieces)
            i_piece = pieces[: len(indent)]
            if self.pieces[-1:] == [""]:
                self.pieces = [pieces[len(indent) :]] + [""]
            elif self.pieces != []:
                self.pieces = [pieces[len(indent) :]]
        self.indent += len(indent)
        for n in node.childNodes:
            if restrict is not None:
                if n.nodeType == n.ELEMENT_NODE and n.tagName in restrict:
                    self.parse(n)
            elif n.nodeType != n.ELEMENT_NODE or n.tagName not in ignore:
                self.parse(n)
        if len(indent) > 0:
            self.pieces = shift(self.pieces, indent, i_piece)
        self.indent -= len(indent)
        old_pieces.extend(self.pieces)
        self.pieces = old_pieces

    def surround_parse(self, node, pre_char, post_char):
        """Parse the subnodes of a given node. Subnodes with tags in the
        `ignore` list are ignored. Prepend `pre_char` and append `post_char` to
        the output in self.pieces."""
        self.add_text(pre_char)
        self.subnode_parse(node)
        self.add_text(post_char)

    # MARK: Helper functions
    def get_specific_subnodes(self, node, name, recursive=0):
        """Given a node and a name, return a list of child `ELEMENT_NODEs`, that
        have a `tagName` matching the `name`. Search recursively for `recursive`
        levels.
        """
        children = [x for x in node.childNodes if x.nodeType == x.ELEMENT_NODE]
        ret = [x for x in children if x.tagName == name]
        if recursive > 0:
            for x in children:
                ret.extend(self.get_specific_subnodes(x, name, recursive - 1))
        return ret

    def get_specific_nodes(self, node, names):
        """Given a node and a sequence of strings in `names`, return a
        dictionary containing the names as keys and child
        `ELEMENT_NODEs`, that have a `tagName` equal to the name.

        """
        nodes = [(x.tagName, x) for x in node.childNodes if x.nodeType == x.ELEMENT_NODE and x.tagName in names]
        return dict(nodes)

    def add_text(self, value):
        """Adds text corresponding to `value` into `self.pieces`."""
        if isinstance(value, (list, tuple)):
            self.pieces.extend(value)
        else:
            self.pieces.append(value)

    def start_new_paragraph(self):
        """Make sure to create an empty line. This is overridden, if the previous
        text ends with the special marker ''. In that case, nothing is done.
        """
        if self.pieces[-1:] == [""]:  # respect special marker
            return
        elif self.pieces == []:  # first paragraph, add '\n', override with ''
            self.pieces = ["\n"]
        elif self.pieces[-1][-1:] != "\n":  # previous line not ended
            self.pieces.extend(["  \n", "\n"])
        else:  # default
            self.pieces.append("\n")

    def add_line_with_subsequent_indent(self, line, indent=4):
        """Add line of text and wrap such that subsequent lines are indented
        by `indent` spaces.
        """
        if isinstance(line, (list, tuple)):
            line = "".join(line)
        line = line.strip()
        width = self.textwidth - self.indent - indent
        wrapped_lines = textwrap.wrap(line[indent:], width=width)
        for i in range(len(wrapped_lines)):
            if wrapped_lines[i] != "":
                wrapped_lines[i] = indent * " " + wrapped_lines[i]
        self.pieces.append(line[:indent] + "\n".join(wrapped_lines)[indent:] + "  \n")

    def extract_text(self, node):
        """Return the string representation of the node or list of nodes by parsing the
        subnodes, but returning the result as a string instead of adding it to `self.pieces`.
        Note that this allows extracting text even if the node is in the ignore list.
        """
        if not isinstance(node, (list, tuple)):
            node = [node]
        pieces, self.pieces = self.pieces, [""]
        for n in node:
            for sn in n.childNodes:
                self.parse(sn)
        ret = "".join(self.pieces)
        self.pieces = pieces
        return ret

    def get_function_signature(self, node):
        """Returns the function signature string for memberdef nodes."""
        name = self.extract_text(self.get_specific_subnodes(node, "name"))
        if self.with_type_info:
            argsstring = self.extract_text(self.get_specific_subnodes(node, "argsstring"))
        else:
            argsstring = []
            param_id = 1
            for n_param in self.get_specific_subnodes(node, "param"):
                declname = self.extract_text(self.get_specific_subnodes(n_param, "declname"))
                if not declname:
                    declname = "arg" + str(param_id)
                defval = self.extract_text(self.get_specific_subnodes(n_param, "defval"))
                if defval:
                    defval = "=" + defval
                argsstring.append(declname + defval)
                param_id = param_id + 1
            argsstring = "(" + ", ".join(argsstring) + ")"
        type = self.extract_text(self.get_specific_subnodes(node, "type"))
        function_definition = name + argsstring
        if type != "" and type != "void":
            function_definition = function_definition + " -> " + type
        return "`" + function_definition + "`  "

    # MARK: Special parsing tasks (need to be called manually)
    def make_constructor_list(self, constructor_nodes, classname):
        """Produces the "Constructors" section and the constructor signatures
        (since swig does not do so for classes) for class docstrings."""
        if constructor_nodes == []:
            return
        self.add_text(["\n", "Constructors", "\n", "------------"])
        for n in constructor_nodes:
            self.add_text("\n")
            self.add_line_with_subsequent_indent("* " + self.get_function_signature(n))
            self.subnode_parse(n, pieces=[], indent=4, ignore=["definition", "name"])

    def make_attribute_list(self, node):
        """Produces the "Attributes" section in class docstrings for public
        member variables (attributes).
        """
        atr_nodes = []
        for n in self.get_specific_subnodes(node, "memberdef", recursive=2):
            if n.attributes["kind"].value == "variable" and n.attributes["prot"].value == "public":
                atr_nodes.append(n)
        if not atr_nodes:
            return
        self.add_text(["\n", "Attributes", "\n", "----------"])
        for n in atr_nodes:
            name = self.extract_text(self.get_specific_subnodes(n, "name"))
            self.add_text(["\n* ", "`", name, "`", " : "])
            self.add_text(["`", self.extract_text(self.get_specific_subnodes(n, "type")), "`"])
            self.add_text("  \n")
            restrict = ["briefdescription", "detaileddescription"]
            self.subnode_parse(n, pieces=[""], indent=4, restrict=restrict)

    def get_memberdef_nodes_and_signatures(self, node, kind):
        """Collects the memberdef nodes and corresponding signatures that
        correspond to public function entries that are at most depth 2 deeper
        than the current (compounddef) node. Returns a dictionary with
        function signatures (what swig expects after the %feature directive)
        as keys, and a list of corresponding memberdef nodes as values."""
        sig_dict = {}
        sig_prefix = ""
        if kind in ("file", "namespace"):
            ns_node = node.getElementsByTagName("innernamespace")
            if not ns_node and kind == "namespace":
                ns_node = node.getElementsByTagName("compoundname")
            if ns_node:
                sig_prefix = self.extract_text(ns_node[0]) + "::"
        elif kind in ("class", "struct"):
            # Get the full function name.
            cn_node = node.getElementsByTagName("compoundname")
            sig_prefix = self.extract_text(cn_node[0]) + "::"

        md_nodes = self.get_specific_subnodes(node, "memberdef", recursive=2)
        for n in md_nodes:
            if n.attributes["prot"].value != "public":
                continue
            if n.attributes["kind"].value in ["variable", "typedef"]:
                continue
            if not self.get_specific_subnodes(n, "definition"):
                continue
            name = self.extract_text(self.get_specific_subnodes(n, "name"))
            if name[:8] == "operator":
                continue
            sig = sig_prefix + name
            if sig in sig_dict:
                sig_dict[sig].append(n)
            else:
                sig_dict[sig] = [n]
        return sig_dict

    def handle_typical_memberdefs_no_overload(self, signature, memberdef_nodes):
        """Produce standard documentation for memberdef_nodes."""
        for n in memberdef_nodes:
            self.add_text(["\n", '%feature("docstring") ', signature, ' "', "\n"])
            if self.with_function_signature:
                self.add_line_with_subsequent_indent(self.get_function_signature(n))
            self.subnode_parse(n, pieces=[], ignore=["definition", "name"])
            self.add_text(['";', "\n"])

    def handle_typical_memberdefs(self, signature, memberdef_nodes):
        """Produces docstring entries containing an "Overloaded function"
        section with the documentation for each overload, if the function is
        overloaded and self.with_overloaded_functions is set. Else, produce
        normal documentation.
        """
        if len(memberdef_nodes) == 1 or not self.with_overloaded_functions:
            self.handle_typical_memberdefs_no_overload(signature, memberdef_nodes)
            return

        self.add_text(["\n", '%feature("docstring") ', signature, ' "', "\n"])
        if self.with_function_signature:
            for n in memberdef_nodes:
                self.add_line_with_subsequent_indent(self.get_function_signature(n))
        self.add_text("\n")
        self.add_text(["Overloaded function", "\n", "-------------------"])
        for n in memberdef_nodes:
            self.add_text("\n")
            self.add_line_with_subsequent_indent("* " + self.get_function_signature(n))
            self.subnode_parse(n, pieces=[], indent=4, ignore=["definition", "name"])
        self.add_text(['";', "\n"])

    # MARK: Tag handlers

    def do_linebreak(self, node):
        self.add_text("  ")

    def do_ndash(self, node):
        self.add_text("--")

    def do_mdash(self, node):
        self.add_text("---")

    def do_emphasis(self, node):
        self.surround_parse(node, "*", "*")

    def do_bold(self, node):
        self.surround_parse(node, "**", "**")

    def do_computeroutput(self, node):
        self.surround_parse(node, "`", "`")

    def do_heading(self, node):
        self.start_new_paragraph()
        pieces, self.pieces = self.pieces, [""]
        level = int(node.attributes["level"].value)
        self.subnode_parse(node)
        if level == 1:
            self.pieces.insert(0, "\n")
            self.add_text(["\n", len("".join(self.pieces).strip()) * "="])
        elif level == 2:
            self.add_text(["\n", len("".join(self.pieces).strip()) * "-"])
        elif level >= 3:
            self.pieces.insert(0, level * "#" + " ")
        # make following text have no gap to the heading:
        pieces.extend(["".join(self.pieces) + "  \n", ""])
        self.pieces = pieces

    def do_verbatim(self, node):
        self.start_new_paragraph()
        self.subnode_parse(node, pieces=[""], indent=4)

    def do_blockquote(self, node):
        self.start_new_paragraph()
        self.subnode_parse(node, pieces=[""], indent="> ")

    def do_hruler(self, node):
        self.start_new_paragraph()
        self.add_text("* * * * *  \n")

    def do_includes(self, node):
        self.add_text("\nC++ includes: ")
        self.subnode_parse(node)
        self.add_text("\n")

    # MARK: Para tag handler
    def do_para(self, node):
        """This is the only place where text wrapping is automatically performed.
        Generally, this function parses the node (locally), wraps the text, and
        then adds the result to self.pieces. However, it may be convenient to
        allow the previous content of self.pieces to be included in the text
        wrapping. For this, use the following *convention*:
        If self.pieces ends with '', treat the _previous_ entry as part of the
        current paragraph. Else, insert new-line and start a new paragraph
        and "wrapping context".
        Paragraphs always end with '  \n', but if the parsed content ends with
        the special symbol '', this is passed on.
        """
        if self.pieces[-1:] == [""]:
            pieces, self.pieces = self.pieces[:-2], self.pieces[-2:-1]
        else:
            self.add_text("\n")
            pieces, self.pieces = self.pieces, [""]
        self.subnode_parse(node)
        dont_end_paragraph = self.pieces[-1:] == [""]
        # Now do the text wrapping:
        width = self.textwidth - self.indent
        wrapped_para = []
        for line in "".join(self.pieces).splitlines():
            keep_markdown_newline = line[-2:] == "  "
            w_line = textwrap.wrap(line, width=width, break_long_words=False)
            if w_line == []:
                w_line = [""]
            if keep_markdown_newline:
                w_line[-1] = w_line[-1] + "  "
            for wl in w_line:
                wrapped_para.append(wl + "\n")
        if wrapped_para:
            if wrapped_para[-1][-3:] != "  \n":
                wrapped_para[-1] = wrapped_para[-1][:-1] + "  \n"
            if dont_end_paragraph:
                wrapped_para.append("")
        pieces.extend(wrapped_para)
        self.pieces = pieces

    # MARK: List tag handlers
    def do_itemizedlist(self, node):
        if self.listitem == "":
            self.start_new_paragraph()
        elif self.pieces != [] and self.pieces[-1:] != [""]:
            self.add_text("\n")
        listitem = self.listitem
        if self.listitem in ["*", "-"]:
            self.listitem = "-"
        else:
            self.listitem = "*"
        self.subnode_parse(node)
        self.listitem = listitem

    def do_orderedlist(self, node):
        if self.listitem == "":
            self.start_new_paragraph()
        elif self.pieces != [] and self.pieces[-1:] != [""]:
            self.add_text("\n")
        listitem = self.listitem
        self.listitem = 0
        self.subnode_parse(node)
        self.listitem = listitem

    def do_listitem(self, node):
        try:
            self.listitem = int(self.listitem) + 1
            item = str(self.listitem) + ". "
        except BaseException:
            item = str(self.listitem) + " "
        self.subnode_parse(node, item, indent=4)

    # MARK: Parameter list tag handlers
    def do_parameterlist(self, node):
        self.start_new_paragraph()
        text = "unknown"
        for key, val in node.attributes.items():
            if key == "kind":
                if val == "param":
                    text = "Parameters"
                elif val == "exception":
                    text = "Exceptions"
                elif val == "retval":
                    text = "Returns"
                else:
                    text = val
                break
        if self.indent == 0:
            self.add_text([text, "\n", len(text) * "-", "\n"])
        else:
            self.add_text([text, ":  \n"])
        self.subnode_parse(node)

    def do_parameteritem(self, node):
        self.subnode_parse(node, pieces=["* ", ""])

    def do_parameternamelist(self, node):
        self.subnode_parse(node)
        self.add_text([" :", "  \n"])

    def do_parametername(self, node):
        if self.pieces != [] and self.pieces != ["* ", ""]:
            self.add_text(", ")
        data = self.extract_text(node)
        self.add_text(["`", data, "`"])

    def do_parameterdescription(self, node):
        self.subnode_parse(node, pieces=[""], indent=4)

    # MARK: Section tag handler
    def do_simplesect(self, node):
        kind = node.attributes["kind"].value
        if kind in ("date", "rcs", "version"):
            return
        self.start_new_paragraph()
        if kind == "warning":
            self.subnode_parse(node, pieces=["**Warning**: ", ""], indent=4)
        elif kind == "see":
            self.subnode_parse(node, pieces=["See also: ", ""], indent=4)
        elif kind == "return":
            if self.indent == 0:
                pieces = ["Returns", "\n", len("Returns") * "-", "\n", ""]
            else:
                pieces = ["Returns:", "\n", ""]
            self.subnode_parse(node, pieces=pieces)
        else:
            self.subnode_parse(node, pieces=[kind + ": ", ""], indent=4)

    # MARK: %feature("docstring") producing tag handlers
    def do_compounddef(self, node):
        """This produces %feature("docstring") entries for classes, and handles
        class, namespace and file memberdef entries specially to allow for
        overloaded functions. For other cases, passes parsing on to standard
        handlers (which may produce unexpected results).
        """
        kind = node.attributes["kind"].value
        if kind in ("class", "struct"):
            prot = node.attributes["prot"].value
            if prot != "public":
                return
            self.add_text("\n\n")
            classdefn = self.extract_text(self.get_specific_subnodes(node, "compoundname"))
            classname = classdefn.split("::")[-1]
            self.add_text(f'%feature("docstring") {classdefn} "\n')

            if self.with_constructor_list:
                constructor_nodes = []
                for n in self.get_specific_subnodes(node, "memberdef", recursive=2):
                    if n.attributes["prot"].value == "public":
                        if (
                            self.extract_text(self.get_specific_subnodes(n, "definition"))
                            == classdefn + "::" + classname
                        ):
                            constructor_nodes.append(n)
                for n in constructor_nodes:
                    self.add_line_with_subsequent_indent(self.get_function_signature(n))

            names = ("briefdescription", "detaileddescription")
            sub_dict = self.get_specific_nodes(node, names)
            for n in ("briefdescription", "detaileddescription"):
                if n in sub_dict:
                    self.parse(sub_dict[n])
            if self.with_constructor_list:
                self.make_constructor_list(constructor_nodes, classname)
            if self.with_attribute_list:
                self.make_attribute_list(node)

            sub_list = self.get_specific_subnodes(node, "includes")
            if sub_list:
                self.parse(sub_list[0])
            self.add_text(['";', "\n"])

            names = ["compoundname", "briefdescription", "detaileddescription", "includes"]
            self.subnode_parse(node, ignore=names)

        elif kind in ("file", "namespace"):
            nodes = node.getElementsByTagName("sectiondef")
            for n in nodes:
                self.parse(n)

        # now explicitely handle possibly overloaded member functions.
        if kind in ["class", "struct", "file", "namespace"]:
            md_nodes = self.get_memberdef_nodes_and_signatures(node, kind)
            for sig in md_nodes:
                self.handle_typical_memberdefs(sig, md_nodes[sig])

    def do_memberdef(self, node):
        """Handle cases outside of class, struct, file or namespace. These are
        now dealt with by `handle_overloaded_memberfunction`.
        Do these even exist???
        """
        prot = node.attributes["prot"].value
        node.attributes["id"].value
        kind = node.attributes["kind"].value
        tmp = node.parentNode.parentNode.parentNode
        compdef = tmp.getElementsByTagName("compounddef")[0]
        cdef_kind = compdef.attributes["kind"].value
        if cdef_kind in ("file", "namespace", "class", "struct"):
            # These cases are now handled by `handle_typical_memberdefs`
            return
        if prot != "public":
            return
        first = self.get_specific_nodes(node, ("definition", "name"))
        name = self.extract_text(first["name"])
        if name[:8] == "operator":  # Don't handle operators yet.
            return
        if "definition" not in first or kind in ["variable", "typedef"]:
            return

        data = self.extract_text(first["definition"])
        self.add_text("\n")
        self.add_text(["/* where did this entry come from??? */", "\n"])
        self.add_text(f'%feature("docstring") {data} "\n{data}')

        for n in node.childNodes:
            if n not in first.values():
                self.parse(n)
        self.add_text(['";', "\n"])

    # MARK: Entry tag handlers (dont print anything meaningful)
    def do_sectiondef(self, node):
        kind = node.attributes["kind"].value
        if kind in ("public-func", "func", "user-defined", ""):
            self.subnode_parse(node)

    def do_header(self, node):
        """For a user defined section def a header field is present
        which should not be printed as such, so we comment it in the
        output."""
        data = self.extract_text(node)
        self.add_text(f"\n/*\n {data} \n*/\n")
        # If our immediate sibling is a 'description' node then we
        # should comment that out also and remove it from the parent
        # node's children.
        parent = node.parentNode
        idx = parent.childNodes.index(node)
        if len(parent.childNodes) >= idx + 2:
            nd = parent.childNodes[idx + 2]
            if nd.nodeName == "description":
                nd = parent.removeChild(nd)
                self.add_text("\n/*")
                self.subnode_parse(nd)
                self.add_text("\n*/\n")

    def do_member(self, node):
        kind = node.attributes["kind"].value
        refid = node.attributes["refid"].value
        if kind == "function" and refid[:9] == "namespace":
            self.subnode_parse(node)

    def do_doxygenindex(self, node):
        self.multi = 1
        comps = node.getElementsByTagName("compound")
        for c in comps:
            refid = c.attributes["refid"].value
            fname = refid + ".xml"
            if not os.path.exists(fname):
                fname = os.path.join(self.my_dir, fname)
            if not self.quiet:
                print(f"parsing file: {fname}")
            p = Doxy2SWIG(
                fname,
                with_function_signature=self.with_function_signature,
                with_type_info=self.with_type_info,
                with_constructor_list=self.with_constructor_list,
                with_attribute_list=self.with_attribute_list,
                with_overloaded_functions=self.with_overloaded_functions,
                textwidth=self.textwidth,
                quiet=self.quiet,
            )
            p.generate()
            self.pieces.extend(p.pieces)


# MARK: main


def main():
    usage = __doc__
    parser = optparse.OptionParser(usage)
    parser.add_option(
        "-f",
        "--function-signature",
        action="store_true",
        default=False,
        dest="f",
        help='include function signature in the documentation. This is handy when not using swig auto-generated function definitions %feature("autodoc", [0,1])',
    )
    parser.add_option(
        "-t",
        "--type-info",
        action="store_true",
        default=False,
        dest="t",
        help="include type information for arguments in function signatures. This is similar to swig autodoc level 1",
    )
    parser.add_option(
        "-c",
        "--constructor-list",
        action="store_true",
        default=False,
        dest="c",
        help="generate a constructor list for class documentation. Useful for target languages where the object construction should be documented in the class documentation.",
    )
    parser.add_option(
        "-a",
        "--attribute-list",
        action="store_true",
        default=False,
        dest="a",
        help="generate an attributes list for class documentation. Useful for target languages where class attributes should be documented in the class documentation.",
    )
    parser.add_option(
        "-o",
        "--overloaded-functions",
        action="store_true",
        default=False,
        dest="o",
        help="collect all documentation for overloaded functions. Useful for target languages that have no concept of overloaded functions, but also to avoid having to attach the correct docstring to each function overload manually",
    )
    parser.add_option(
        "-w",
        "--width",
        type="int",
        action="store",
        dest="w",
        default=80,
        help="textwidth for wrapping (default: 80). Note that the generated lines may include 2 additional spaces (for markdown).",
    )
    parser.add_option(
        "-q", "--quiet", action="store_true", default=False, dest="q", help="be quiet and minimize output"
    )

    options, args = parser.parse_args()
    if len(args) != 2:
        parser.error("no input and output specified")

    p = Doxy2SWIG(
        args[0],
        with_function_signature=options.f,
        with_type_info=options.t,
        with_constructor_list=options.c,
        with_attribute_list=options.a,
        with_overloaded_functions=options.o,
        textwidth=options.w,
        quiet=options.q,
    )
    p.generate()
    p.write(args[1])


if __name__ == "__main__":
    main()



"""Orthogonal Array package test functions"""

import io
import tempfile
import unittest
import unittest.mock as mock

import numpy as np
import oalib

import oapackage
import oapackage.Doptim
import oapackage.scanf


class TestDoptimize(unittest.TestCase):
    def setUp(self):
        self.arrayclass = oapackage.arraydata_t(2, 16, 0, 6)
        self.dds = np.random.rand(20, 3)
        self.dds2 = np.array([[1, 1, 1], [1, 2, 1], [1, 2, 3], [2, 0, 1]])

        self.guitest = True

    def test_Doptimize_large_number_of_runs(self):
        arrayclass = oapackage.arraydata_t(2, 1000, 0, 3)
        scores, dds, sols, _ = oapackage.Doptim.Doptimize(
            arrayclass, nrestarts=10, optimfunc=[1, 0, 0], verbose=0, maxtime=9, niter=10, dverbose=0
        )

        with self.assertRaises(Exception):
            arrayclass = oapackage.arraydata_t(2, 100_000, 0, 3)
            scores, dds, sols, _ = oapackage.Doptim.Doptimize(
                arrayclass, nrestarts=10, optimfunc=[1, 0, 0], verbose=0, maxtime=9, niter=10
            )

    def test_custom_optim(self):
        def optimfunc(x):
            return x[0] + x[1] + x[2]

        with mock.patch("sys.stdout", new_callable=io.StringIO):
            scores, dds, sols, n = oapackage.Doptim.Doptimize(
                self.arrayclass,
                nrestarts=2,
                optimfunc=optimfunc,
                verbose=1,
                maxtime=18,
                selectpareto=False,
                nout=None,
                method=oalib.DOPTIM_UPDATE,
                niter=1000,
                nabort=0,
                dverbose=0,
            )
        self.assertEqual(len(scores), n)
        self.assertEqual(len(dds), n)
        self.assertEqual(len(sols), n)
        scores, dds, sols, n = oapackage.Doptim.Doptimize(
            self.arrayclass,
            nrestarts=2,
            optimfunc=None,
            verbose=0,
            maxtime=6,
            selectpareto=False,
            nout=None,
            method=oalib.DOPTIM_UPDATE,
            niter=30,
            nabort=0,
            dverbose=0,
        )

    def test_Doptimize_nonzero_strength(self):
        arrayclass = oapackage.arraydata_t(2, 16, 2, 6)
        with self.assertWarns(UserWarning):
            scores, dds, sols, _ = oapackage.Doptim.Doptimize(
                arrayclass, nrestarts=1, verbose=0, maxtime=9, selectpareto=False, nout=None, niter=1000, dverbose=0
            )

    def test_Doptimize_selectDn(self):
        scores, dds, sols, _ = oapackage.Doptim.Doptimize(
            self.arrayclass,
            nrestarts=10,
            optimfunc=[1, 0, 0],
            verbose=0,
            maxtime=9,
            selectpareto=False,
            nout=None,
            method=oalib.DOPTIM_UPDATE,
            niter=1000,
            nabort=0,
            dverbose=0,
        )

        result = oapackage.Doptim.selectDn(scores, dds, sols, nout=1, sortfull=True)
        self.assertTrue(len(result[2]) == 1)

    def test_optimDeffPython(self):
        al = oapackage.exampleArray(2, 0)
        with mock.patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            _, al = oapackage.Doptim.optimDeffPython(
                al, arrayclass=None, niter=1000, nabort=1500, verbose=1, alpha=[1, 0, 0], method=0
            )
            stdout = mock_stdout.getvalue()
            self.assertTrue(stdout.startswith("optimDeff: initial Deff 0.0000"))

        for method in [oapackage.oalib.DOPTIM_SWAP, oapackage.oalib.DOPTIM_FLIP, oapackage.oalib.DOPTIM_UPDATE]:
            r, al = oapackage.Doptim.optimDeffPython(
                al, arrayclass=None, niter=100, nabort=200, verbose=0, alpha=[1, 0, 0], method=method
            )

    def test_generateDscatter(self):
        if self.guitest:
            fig = 100
            with mock.patch("sys.stdout", new_callable=io.StringIO):
                oapackage.Doptim.generateDscatter(
                    self.dds,
                    second_index=0,
                    first_index=1,
                    lbls=None,
                    verbose=1,
                    ndata=3,
                    nofig=True,
                    fig=fig,
                    scatterarea=80,
                )

    def test_generateDpage(self):
        outputdir = tempfile.mkdtemp()
        allarrays = [oapackage.exampleArray(2, 0), oapackage.exampleArray(2, 0)]
        dds = np.array([A.Defficiencies() for A in allarrays])
        arrayclass = oapackage.arraylink2arraydata(allarrays[0])

        with mock.patch("sys.stdout", new_callable=io.StringIO):
            page = oapackage.Doptim.generateDpage(
                outputdir, arrayclass, dds, allarrays, fig=None, optimfunc=[1, 0, 0], nofig=True
            )
        if self.guitest and 0:
            print("test_generateDpage: run gui test")
            page = oapackage.Doptim.generateDpage(
                outputdir, arrayclass, dds, allarrays, fig=100, optimfunc=[1, 0, 0], nofig=True
            )

            self.assertIsInstance(page, oapackage.markup.page)
            try:
                import matplotlib

                matplotlib.pyplot.close(100)
            except ImportError:
                pass

    def test_filterPareto(self):
        dds = self.dds2
        scores = np.arange(dds.shape[0])
        sols = [None] * scores.size
        s, _, _ = oapackage.Doptim.filterPareto(scores, dds, sols, verbose=0)
        self.assertEqual(list(s), [2, 3])

    def test_calcScore(self):
        dds = np.random.rand(10, 3)
        scores = oapackage.Doptim.calcScore(dds, optimfunc=[1, 2, 3])
        self.assertEqual(scores.shape, (dds.shape[0],))

    def test_array2Dtable(self):
        sols = [oapackage.exampleArray(9, 0)]
        _ = oapackage.Doptim.array2Dtable(sols, verbose=1, titlestr=None)


if __name__ == "__main__":
    """Test code"""
    unittest.main()
    # t=TestDoptimize()



import unittest

import numpy as np
from numpy.random import MT19937, RandomState, SeedSequence

import oapackage


def reduce(graph, colors):
    tr = oapackage.reduceGraphNauty(graph, colors=colors, verbose=0)
    tri = inverse_permutation(tr)
    graph_reduced = oapackage.transformGraphMatrix(graph, tri)
    colors_reduced = [colors[idx] for idx in tr]
    return graph_reduced, colors_reduced, tr


def inverse_permutation(perm):
    inverse = [0] * len(perm)
    for i, p in enumerate(perm):
        inverse[p] = i
    return inverse


class TestGWLP(unittest.TestCase):
    def test_reduceNauty_directed_graphs(self):
        rs = RandomState(MT19937(SeedSequence(123456789)))

        for jj, graph_size in enumerate([2, 4, 5, 6, 8]):
            graph = (rs.random((graph_size, graph_size)) > 0.5).astype(int)
            colors = [int(v) for v in rs.random(graph_size) > 0.5]

            graph_reduced, colors_reduced, tr = reduce(graph, colors)

            for ii in range(4):
                perm = rs.permutation(graph_size)

                graph2 = graph[perm, :][:, perm]
                colors2 = [colors[idx] for idx in perm]
                graph2_reduced, colors2_reduced, tr2 = reduce(graph2, colors2)

                self.assertTrue(np.all(graph_reduced == graph2_reduced))
                self.assertTrue(np.all(colors_reduced == colors2_reduced))


if __name__ == "__main__":
    """ Test code """
    unittest.main()
